@@ -1,6 +1,11 @@
-{-# LANGUAGE ScopedTypeVariables, TypeFamilies, ParallelListComp, CPP, BangPatterns #-}
+{-# LANGUAGE ScopedTypeVariables, TypeFamilies, ParallelListComp, CPP, BangPatterns, TemplateHaskell #-}
 module Main where
 
+import Paths_timeplot (version)
+import Data.Version (showVersion)
+import Distribution.VcsRevision.Git
+import Language.Haskell.TH.Syntax
+
 import Control.Monad
 import Data.List
 import Data.Ord
@@ -16,10 +21,6 @@ import Data.Accessor
 
 import Graphics.Rendering.Chart
 
-#ifdef HAVE_GTK
-import Graphics.Rendering.Chart.Gtk
-#endif
-
 import Tools.TimePlot.Types
 import Tools.TimePlot.Conf
 import Tools.TimePlot.Source
@@ -37,26 +38,27 @@ import qualified Tools.TimePlot.Incremental as I
 -- Pass 2:
 --  * Generate plot data (one-pass multiplexed to tracks)
 -- 
-makeChart :: (S.ByteString -> [ChartKind UTCTime]) -> 
-             IO [(UTCTime, InEvent)] ->
-             Maybe UTCTime -> Maybe UTCTime ->
-             (UTCTime -> String -> String) -> 
+makeChart :: (S.ByteString -> [(ChartKind LocalTime, S.ByteString)]) -> 
+             IO [(LocalTime, InEvent)] ->
+             Maybe LocalTime -> Maybe LocalTime ->
+             (LocalTime -> String -> String) -> 
              IO (Renderable ())
 makeChart chartKindF readEvents minT maxT transformLabel = do
   events <- readEvents
   if null events
     then return emptyRenderable
     else do
+      let dropLateEvents  es = case maxT of Just t -> takeWhile ((<t) . fst) es ; Nothing -> es
       -- Pass 1: find out min/max time and final track names.
-      let i2o t KindNone                = []
-          i2o t (KindWithin mapName sk) = [(mapName t, sk)]
-          i2o t k                       = [(t, k)]
+      let i2o t (KindNone,                _) = []
+          i2o t (KindWithin mapName sk, suf) = [(S.append (mapName t) suf, sk)]
+          i2o t (k,                     suf) = [(S.append t suf, k)]
       let i2oTracks t = concatMap (i2o t) (chartKindF t)
       let t0 = fst (head events)
       let (minTime, maxTime, outTracks) = foldl' 
             (\(!mi,!ma,!ts) (t,e) -> (min t mi, max t ma, foldr (uncurry M.insert) ts (i2oTracks $ evt_track e))) 
             (t0, t0, M.empty) 
-            events
+            (dropLateEvents events)
 
       let minOutTime = case minT of Just t -> t ; Nothing -> minTime
       let maxOutTime = case maxT of Just t -> t ; Nothing -> maxTime
@@ -64,34 +66,26 @@ makeChart chartKindF readEvents minT maxT transformLabel = do
       let commonTimeAxis = transformLabels $ autoAxis [minOutTime, maxOutTime]
       
       -- Pass 2
-      events' <- readEvents
+      events' <- dropLateEvents `fmap` readEvents
       let eventsToTracks = [(outTrack, (t,e)) | (t,e) <- events', (outTrack,_) <- i2oTracks (evt_track e)]
 
       let initPlot track = initGen (outTracks M.! track) (S.unpack track) minTime maxTime
       let plots = I.runStreamSummary (I.byKey initPlot) eventsToTracks
       
       -- Render
-      return $ renderLayout1sStacked $ map (dataToPlot commonTimeAxis) (M.elems plots)
+      return $ renderLayout1sStacked $ map (dataToPlot commonTimeAxis (minOutTime,maxOutTime)) (M.elems plots)
 
 showHelp = mapM_ putStrLn [ "",
   "tplot - a tool for drawing timing diagrams.",
   "        See http://www.haskell.org/haskellwiki/Timeplot",
-#ifdef HAVE_GTK  
-  "Usage: tplot [-o OFILE] [-of {png|pdf|ps|svg|x}] [-or 640x480]",
-  "             -if IFILE [-tf TF] ",
-  "             [{+|-}k Pat1 Kind1 {+|-}k Pat2 Kind2 ...] [{+|-}dk KindN]",
-  "             [-fromTime TIME] [-toTime TIME] [-baseTime TIME]",
-  "  -o  OFILE  - output file (required if -of is not x)",
-  "  -of        - output format (x means draw result in a window, default:",
-  "               extension of -o)",
-#else
   "Usage: tplot [-o OFILE] [-of {png|pdf|ps|svg}] [-or 640x480]",
   "             -if IFILE [-tf TF] ",
-  "             [{+|-}k Pat1 Kind1 {+|-}k Pat2 Kind2 ...] [{+|-}dk KindN]",
+  "             [{+|-}k Pat1 '[+Suf1] Kind1' {+|-}k Pat2 '[+Suf2] Kind2' ...] [{+|-}dk '[+Suf] KindN']",
   "             [-fromTime TIME] [-toTime TIME] [-baseTime TIME]",
+  "  --version  - show version information",
+  "  --help     - show this help",
   "  -o  OFILE  - output file",
   "  -of        - output format (default: extension of -o)",
-#endif
   "  -or        - output resolution (default 640x480)",
   "  -if IFILE  - input file; '-' means 'read from stdin'",
   "               NOTE: for large datasets, use actual files, not stdin,",
@@ -105,6 +99,7 @@ showHelp = mapM_ putStrLn [ "",
   "               We also support %OS for fractional seconds (i.e. %OS will",
   "               parse 12.4039 or 12,4039) and %^[+-][N]s for ten-powers ",
   "               of seconds since epoch, for example %^-3s is ms since epoch.",
+  "               -tf elapsed means times are fractional seconds since an unknown moment.",
   "               Default: 'date %Y-%m-%d %H:%M:%OS'",
   "  {+|-}dk    - set default diagram kind",
   "  {+|-}k P K - set diagram kind for tracks matching regex P (in the format",
@@ -115,6 +110,15 @@ showHelp = mapM_ putStrLn [ "",
   "               matched: a track is drawn acc. to all matching +k, to +dk",
   "               AND ALSO to the first matching -k, or -dk if none of -k",
   "               match",
+  "               EXPLANATION OF SUF:",
+  "               If '+Suf' is present (e.g. +k request '+frequency freq 60'),",
+  "               then '.Suf' is appended to the input track name while mapping it",
+  "               to the output track. This is so that a single input track can",
+  "               participate in many output tracks - tplot can have only one output",
+  "               track with a particular name, so you can't have one track named",
+  "               'request' which draws 'freq 60' and another one for 'hist 60' - ",
+  "               you should use +k request '+frequency freq 60' +k request '+histogram hist 60'",
+  "               and you'll get output tracks 'request.frequency' and 'request.histogram'",
   "  -fromTime  - filter records whose time is >= this time",
   "               (formatted according to -tf)",
   "  -toTime    - filter records whose time is <  this time",
@@ -138,12 +142,21 @@ showHelp = mapM_ putStrLn [ "",
   "  'none' - do not plot this track",
   "  'event' is for event diagrams: activities are drawn like --[===]--- ,",
   "     pulse events like --|-- with a label over '|'",
-  "  'duration XXXX' - plot any kind of diagram over the *durations* of events",
+  "  'duration [drop] XXXX' - plot any kind of diagram over the *durations* of events",
   "     on a track (delimited by > ... <), for example 'duration quantile",
   "     300 0.25,0.5,0.75' will plot these quantiles of durations of the",
   "     events. This is useful where your log looks like 'Started processing'",
   "     ... 'Finished processing': you can plot processing durations without",
   "     computing them yourself. Very useful inside 'within'!",
+  "     If you use 'drop', then names of the original input tracks will be dropped",
+  "     before putting the events onto the output track, e.g. an event rtime.14e3ac1",
+  "     when used by 'within[.] duration drop dots', will be put onto the output track",
+  "     'rtime', with input track 'rtime'. When used by 'within[.] duration dots',",
+  "     its input track will still be rtime.14e3ac1. The difference is whether",
+  "     the output of 'duration' appears to 'XXXX' as a single or multiple input tracks.",
+  "     E.g. if you're measuring durations of processing unique requests with rtime.REQID,",
+  "     then use 'drop'; if it's durations of processing at certain stages with rtime.STAGE",
+  "     then don't.",
   "  'within[C] XXXX' - draw plot XXXX over events grouped by their track's name ",
   "     before separator C. For example, if you have processes",
   "     named 'MACHINE-PID' (i.e. UNIT027-8532) say 'begin something' / ",
@@ -203,21 +216,26 @@ showHelp = mapM_ putStrLn [ "",
   "     way as in cumsum."
   ]
 
+showGitVersion = $(do
+  v <- qRunIO getRevision
+  lift $ case v of
+    Nothing           -> "<none>"
+    Just (hash,True)  -> hash ++ " (with local modifications)"
+    Just (hash,False) -> hash)
 
 main = do
   args <- getArgs
   mainWithArgs args
 mainWithArgs args = do
   when (null args || args == ["--help"]) $ showHelp >> exitSuccess
+  when (null args || args == ["--version"]) $ do
+    putStrLn ("This is timeplot-" ++ showVersion version ++ " (git " ++ showGitVersion ++ ")") >> exitSuccess
   let !conf = readConf args
   let render = case (outFormat conf) of {
       PNG    -> \c w h f -> const () `fmap` renderableToPNGFile c w h f;
       PDF    -> renderableToPDFFile ;
       PS     -> renderableToPSFile  ;
       SVG    -> renderableToSVGFile ;
-#ifdef HAVE_GTK          
-      Window -> \c w h f -> renderableToWindow c w h
-#endif          
     }
   case conf of
     ConcreteConf {
